include_guard(GLOBAL)

if (NOT EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/sdk)
	file(MAKE_DIRECTORY ${CMAKE_CURRENT_SOURCE_DIR}/sdk)
endif()
message("--    SDK_TYPE : $ENV{SDK_TYPE}")
message("-- SDK_INCLUDE : $ENV{SDK_INCLUDE}")
message("--     SDK_LIB : $ENV{SDK_LIB}")

option(ENABLE_FL_DRIVER "Build & Link with FL Drivers!" ON)
option(ENABLE_SEGGER_RTT "Build & Link with SEGGER RTT support, J-Link required!" OFF)
set(TARGET_CPU_FLOAT_ABI soft CACHE STRING "Select float ABI: none, soft, softfp, hard")
set(TARGET_SPECS nano CACHE STRING "Select specs: none, nosys, nano, rdimon, rdimon-v2m")
set(TARGET_HSE_VALUE auto CACHE STRING "Select HSE Speed: auto, <number>")
set(TARGET_HSI_VALUE auto CACHE STRING "Select HSI Speed: auto, <number>")

if ("${TARGET_CPU}" STREQUAL "")
	message(FATAL_ERROR "TARGET_CPU was not specified!")
endif()

string(TOLOWER "${TARGET_CPU}" TARGET_CPU)

if ("${TARGET_CPU}" MATCHES "^fm33l([0])([0-9][0-9])..$")
	if ("${TARGET_CPU}" MATCHES "^fm33l([ecfg])0([124])x$")
		set(TARGET_CPU_TYPE_DEF "FM33L${CMAKE_MATCH_1}0${CMAKE_MATCH_2}x")
		set(TARGET_CPU_TYPE_COMMON "FM33L${CMAKE_MATCH_1}xx")
		set(TARGET_CPU_TYPE cortex-m0)
		set(TARGET_ARM_MATH_TYPE ARM_MATH_CM0)
	else()
		message(FATAL_ERROR "Unsupported CPU[${TARGET_CPU}]!")
	endif()
else()
	message(FATAL_ERROR "Unsupported device [${TARGET_CPU}]!")
endif()

# if ("${TARGET_CPU}" MATCHES "^stm32(f[12]|l[1]).*")
# 	set(TARGET_CPU_TYPE cortex-m3)
# 	set(TARGET_ARM_MATH_TYPE ARM_MATH_CM3)
# elseif("${TARGET_CPU}" MATCHES "^stm32(f[34]|g[4]|l[4]).*")
# 	set(TARGET_CPU_TYPE cortex-m4)
# 	set(TARGET_ARM_MATH_TYPE ARM_MATH_CM4)
# elseif("${TARGET_CPU}" MATCHES "^stm32(f[7]|h[7]).*")
# 	set(TARGET_CPU_TYPE cortex-m7)
# 	set(TARGET_ARM_MATH_TYPE ARM_MATH_CM7)
# elseif("${TARGET_CPU}" MATCHES "^stm32(h[5]|l[5]|u[5]).*")
# 	set(TARGET_CPU_TYPE cortex-m33)
# 	set(TARGET_ARM_MATH_TYPE ARM_MATH_ARMV8MBL)
# elseif("${TARGET_CPU}" MATCHES "^stm32(g[0]|c[0]|l[0]).*")
# 	set(TARGET_CPU_TYPE cortex-m0plus)
# 	set(TARGET_ARM_MATH_TYPE ARM_MATH_CM0PLUS)
# elseif("${TARGET_CPU}" MATCHES "^stm32mp1.*")
# 	set(TARGET_CPU_TYPE cortex-a7)
# endif()
message("--      Device : " ${TARGET_CPU})
message("--         CPU : " ${TARGET_CPU_TYPE})


if(CLANG)
	add_compile_options(--target=arm-none-eabi -Wno-unused-command-line-argument)
	add_link_options(--target=arm-none-eabi -Wno-unused-command-line-argument)
endif()

add_compile_definitions(${TARGET_CPU_TYPE_COMMON} ${TARGET_CPU_TYPE_DEF})
add_compile_options(
	-mthumb
	-mcpu=${TARGET_CPU_TYPE}
	-Wfatal-errors
	-fno-exceptions
	-fno-unwind-tables
	-fno-stack-check
)

set(CMAKE_CXX_FLAGS	"${CMAKE_CXX_FLAGS} -fno-rtti -fno-exceptions -fuse-cxa-atexit")
add_link_options(
	-mthumb
	-mcpu=${TARGET_CPU_TYPE}
	-Wl,--print-memory-usage
	-Wl,--warn-once
	${EMBEDED_GCC_EXTRA_LD_FLAGS}
	-L"$ENV{SDK_LIB}"
)

message("--       Specs : " ${TARGET_SPECS})
select_embeded_spec(${TARGET_SPECS})

include_directories(${CMAKE_CURRENT_SOURCE_DIR}/inc $ENV{SDK_INCLUDE})

if ("${TARGET_CPU_FLOAT_ABI}" MATCHES "soft|softfp|hard")
	add_compile_options(-mfloat-abi=${TARGET_CPU_FLOAT_ABI})
	add_link_options(-mfloat-abi=${TARGET_CPU_FLOAT_ABI})
	message("--   Float ABI : " ${TARGET_CPU_FLOAT_ABI})
elseif(NOT "${TARGET_CPU_FLOAT_ABI}" STREQUAL "none")
	message("${_WARNING_PREFIX} : Invalid Float ABI ${TARGET_CPU_FLOAT_ABI}")
endif()

if ("${TARGET_HSE_VALUE}" STREQUAL "auto")
	message("--   HSE_VALUE : auto")
elseif("${TARGET_HSE_VALUE}" MATCHES "^[0-9]+$")
	add_compile_definitions(HSE_VALUE=${TARGET_HSE_VALUE})
	message("--   HSE_VALUE : " ${TARGET_HSE_VALUE})
else()
	message("${_WARNING_PREFIX} : Invalid HSE_VALUE ${TARGET_HSE_VALUE}")
endif()

if ("${TARGET_HSI_VALUE}" STREQUAL "auto")
	message("--   HSI_VALUE : auto")
elseif("${TARGET_HSI_VALUE}" MATCHES "^[0-9]+$")
	add_compile_definitions(HSI_VALUE=${TARGET_HSI_VALUE})
	message("--   HSI_VALUE : " ${TARGET_HSI_VALUE})
else()
	message("${_WARNING_PREFIX} : Invalid HSI_VALUE ${TARGET_HSI_VALUE}")
endif()

set(STM32CUBE_SDK)

if (NOT EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/sdk/${TARGET_CPU_TYPE_DEF}_FLASH.ld)
	if (EXISTS $ENV{SDK_CMSIS_DEVICES_FL_SRC}/Templates/gcc/linker/${TARGET_CPU_TYPE_DEF}_FLASH.ld)
		file(COPY_FILE
			$ENV{SDK_CMSIS_DEVICES_FL_SRC}/Templates/gcc/linker/${TARGET_CPU_TYPE_DEF}_FLASH.ld
			${CMAKE_CURRENT_SOURCE_DIR}/sdk/${TARGET_CPU_TYPE_DEF}_FLASH.ld
		)
	endif()
endif()
# copy system_xxxx.c
if (NOT EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/sdk/system_${TARGET_CPU_TYPE_COMMON}.c)
	file(COPY_FILE
		$ENV{SDK_CMSIS_DEVICES_FL_SRC}/Templates/system_${TARGET_CPU_TYPE_COMMON}.c
		${CMAKE_CURRENT_SOURCE_DIR}/sdk/system_${TARGET_CPU_TYPE_COMMON}.c
	)
endif()

# copy startup_xxxx.s
if (NOT EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/sdk/startup_${TARGET_CPU_TYPE_COMMON}.s)
	file(COPY_FILE
		$ENV{SDK_CMSIS_DEVICES_FL_SRC}/Templates/GCC/startup_${TARGET_CPU_TYPE_COMMON}.s
		${CMAKE_CURRENT_SOURCE_DIR}/sdk/startup_${TARGET_CPU_TYPE_COMMON}.s
	)
endif()

if (ENABLE_FL_DRIVER)
	message("--   FL Driver : enabled")
	file(GLOB_RECURSE FL_DRIVER_SOURCES
		LIST_DIRECTORIES false
		$ENV{SDK_FL_DRIVER_SRC}/*_fl_*.c
	)
	list(TRANSFORM FL_DRIVER_SOURCES REPLACE "\\\\" "/")
	add_library(ll-driver STATIC
		${FL_DRIVER_SOURCES}
	)
	list(APPEND FM33_SDK fl-driver)
else()
	message("--   FL Driver : disabled")
endif()

if(EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/SDK/${TARGET_CPU_TYPE_COMMON}_FLASH.ld)
	add_link_options(-Wl,-T,${CMAKE_CURRENT_SOURCE_DIR}/SDK/${TARGET_CPU_TYPE_COMMON}_FLASH.ld)
	message("--   LD Script : sdk/${TARGET_CPU_TYPE_COMMON}_FLASH.ld")
else()
	message("${_WARNING_PREFIX} : LD script not found!")
endif()

include(${CMAKE_CURRENT_LIST_DIR}/segger-rtt.cmake)

file(GLOB_RECURSE SOURCE_FILES
	LIST_DIRECTORIES false
	${SEGGER_RTT_SOURCES}
	sdk/*.cxx
	sdk/*.cpp
	sdk/*.cc
	sdk/*.c
	sdk/*.s
	src/*.cxx
	src/*.cpp
	src/*.cc
	src/*.c
	src/*.s
	drivers/*.s
	drivers/*.c
	drivers/*.cxx
	drivers/*.cpp
	drivers/*.cc
)
