JEDI_LANGUAGE_SERVER = utils:search_env("jedi-language-server.exe", "PATH");
-- D_LANGUAGE_SERVER = utils:file_time([[D:\dmd2\windows\bin\serve-d.exe]]) and [[D:\dmd2\windows\bin\serve-d.exe]];

local function _init_lspc(lspc)
	lspc.file_size_limit = 200000;
	function lspc:on_locate_file(file_name, begin_line, begin_col, end_line, end_col)
		mgr:on_locate_file(1, file_name, begin_line, begin_col, end_line, end_col);
	end;
	function lspc:on_diagnostic(src, code, file, begin_line, begin_col, end_line, end_col, severity, info)
		if src == "LSP" and begin_line == 0 and end_line == 0 and info == "" then
			if code == "begin" then
				mgr:clear_doc_diagnostic("diag", file);
			elseif code == "end" then
				-- refresh the current document
				mgr:refresh_doc_diagnostic(file);
			end;
		else
			-- print(file,":", begin_line, "," , begin_col, ":", end_line, ",", end_col, ": diag ", info);
			mgr:add_doc_diagnostic("diag", file, begin_line, begin_col, end_line, end_col, severity, info);
		end;
	end;
	function lspc:on_terminated(code)
		print(string.format("LSP service terminated(%d)!", code));
		show_notify(LANG("msg/lsp/terminated"));
	end;
	function lspc:on_refresh_semantic_tokens(scope, info)
		-- print("on_refresh_semantic_tokens ", scope);
		local all_docs = {mgr.get_all_documents()};
		for i, doc in ipairs(all_docs) do
			lspc:fetch_semantic_tokens(doc, "open");
		end;
	end;
	function lspc:on_document_changed(file, ver)
		local doc = mgr:get_document(file);
		if doc == nil then
			return;
		end;
		lspc:fetch_semantic_tokens(doc, "delta");
		lspc:find_inactive_regions(doc);
	end;
	function lspc:on_update_inactive_regions(file, regions)
		mgr:update_inactive_regions(file, regions);
	end;

	function lspc:on_rename_symbol(changes)
		-- print("on_rename_symbol", changes);
		if changes == nil or #changes < 1 then
			mgr:popup_info_tip("* "..LANG("notify/error/lsp/can_not_rename_symbol").."", 1+0x1000+0x8000);
			return;
		end;
		local doc = mgr.current_document;
		if doc == nil then
			return;
		end;
		local cur_doc_name = doc.file_name;
		local pending_changes = {};
		local pending_changes_count = 0;
		for idx, change in ipairs(changes) do
			local doc_name = table.remove(change, 1);
			if doc_name == cur_doc_name then
				local cursor_line = doc.cursor_line;
				local cursor_col = doc.cursor_column;
				mgr:lock_drawing();
				doc:delete_marked_text_by_tag(marked_text_tag.inplace_suggestion);
				doc:clear_marked_text();
				for idx, op in ipairs(change) do
					local flags = marked_text_flag.renaming + marked_text_flag.placehold_end;
					if (op.begin_line < cursor_line or (op.begin_line == cursor_line and op.begin_col <= cursor_col)) and
						(op.end_line > cursor_line or (op.end_line == cursor_line and op.end_col >= cursor_col)) then
						doc:select_text(op.begin_line, op.begin_col, op.end_line, op.end_col);
					end;
					local id = doc:create_marked_text(op.begin_line, op.begin_col, op.end_line, op.end_col, flags);
					doc:marked_text_set_tag(id, marked_text_tag.rename);
				end;
				mgr:unlock_drawing();
			else
				pending_changes[doc_name] = change;
				pending_changes_count = pending_changes_count + 1;
			end;
		end;
		if pending_changes_count > 0 then
			edx._pending_rename_ops = pending_changes;
		else
			edx._pending_rename_ops = nil;
		end;
	end;
	return lspc;
end;

local LSPC_NO_EMPTY_COMMAND = 0;
local LSPC_SEND_EMPTY_COMMAND = 1;
local LSPC_SEND_EMPTY_NOTIFY = 2;

lsp = {};
function lsp:_init_clangd(cmd)
	local clangd_path = CLANGD_PATH;
	local lspc = _init_lspc(edx:create_lsp_client());
	edx._lsp_clients["cpp"] = lspc;
	local build_path = cmd.build_path;
	local toolset = cmd.toolset;
	local tools = (toolset and toolset.tools) or {};
	local query_driver;
	local lsp_env = {
		["CPATH"] = table.concat(table.iclone(toolset.include, cmd.cmake_inc_path), ";");
		["PATH"] = table.concat(table.iclone(toolset.bin, toolset.lib, {os.getenv("PATH")}), ";");
	};
	if tools.clangd then
		clangd_path = tools.clangd;
	end;
	if toolset.query_driver then
		query_driver = toolset:query_driver(cmd);
	else
		local compiler_path = tools["cxx"] or tools["cc"];
		if toolset["type"] ~= "msvc" and compiler_path then
			compiler_path = compiler_path:match("(.*)(;.*)") or compiler_path;
			if utils:file_time(compiler_path) then
				query_driver = compiler_path;
			end;
		end;
		if query_driver then
		elseif tools["clang"] then
			query_driver = make_path(toolset.home, "**\\clang++*");
		elseif tools["gcc"] then
			query_driver = make_path(toolset.home, "**\\g++*");
		end;
	end;
	lspc.send_empty_command = LSPC_NO_EMPTY_COMMAND;
	local param_compile_commands_dir = "";
	local param_query_driver = "";
	if build_path then
		if #build_path > #cmd.project_path+2 then
			param_compile_commands_dir = [[--compile-commands-dir="]]..build_path:sub(#cmd.project_path+2)..[["]];
		else
			param_compile_commands_dir = [[--compile-commands-dir="]]..build_path..[["]];
		end;
	end;
	if query_driver then
		param_query_driver = [[ --query-driver="]]..query_driver..[["]];
	end;

	local test_compile_commands_json = function(file_path)
		if project:is_ready() and utils:file_time(file_path) == nil then
			print(LANG("msg/clangd/compile_commands_json_not_found"));
		end;
	end;
	local org_file_name = make_path(build_path, "/compile_commands.json");
	test_compile_commands_json(org_file_name);
	if toolset.wsl then
		-- remap wsl path for compile_commands.json
		local query_driver_src = make_path(edx.app_path, "wsl-query-driver.exe");
		local new_driver_time = utils:file_time(query_driver_src);
		if new_driver_time == nil then
			query_driver_src = make_path(edx.startup_path, "wsl-query-driver.exe");
			new_driver_time = utils:file_time(query_driver_src);
		end;
		local query_driver = make_path(build_path, "../wsl-"..toolset.wsl.."-gcc.exe");
		local old_driver_time = utils:file_time(query_driver);
		if old_driver_time == nil or new_driver_time > old_driver_time then
			if old_driver_time ~= nil then
				os.remove(query_driver);
			end;
			print(LANG("msg/clangd/install_query_driver"), query_driver);
			utils:copy_file(query_driver_src, query_driver);
		end;
		query_driver = query_driver:gsub("\\", "/");
		param_query_driver = [[ --query-driver="]]..query_driver..[["]];
	elseif toolset.remote then
		-- remap host path for compile_commands.json
		local query_driver_src = make_path(edx.app_path, "wsl-query-driver.exe");
		local new_driver_time = utils:file_time(query_driver_src);
		if new_driver_time == nil then
			query_driver_src = make_path(edx.startup_path, "wsl-query-driver.exe");
			new_driver_time = utils:file_time(query_driver_src);
		end;
		local query_driver = make_path(build_path, "../ssh-gcc.exe");
		local old_driver_time = utils:file_time(query_driver);
		if old_driver_time == nil or new_driver_time > old_driver_time then
			if old_driver_time ~= nil then
				os.remove(query_driver);
			end;
			print(LANG("msg/clangd/install_query_driver"), query_driver);
			utils:copy_file(query_driver_src, query_driver);
		end;
		query_driver = query_driver:gsub("\\", "/");
		local unmap_path = toolset.unmap_path or function(path) return path; end;
		dbg_call(function()
				local query_driver_target = make_path(build_path, "../target.txt");
				target_file = io.u8_open(query_driver_target, "w+");
				if target_file == nil then
					return;
				end;
				local local_include_path = {};
				for idx, path in ipairs(toolset.include) do
					table.insert(local_include_path, unmap_path(path));
				end;
				target_file:write(
					"Target: "..toolset.target.."\n"..
					"Thread model: posix\n"..
					"Supported LTO compression algorithms: zlib\n"..
					"gcc version "..toolset.version.."\n"..
					"#include \"...\" search starts here:\n"..
					"#include <...> search starts here:\n"..
					" "..
					table.concat(local_include_path, "\n ")..
					"\n"..
					"End of search list.\n\n"
				);
				target_file:close();
		end, "");
		param_query_driver = [[ --query-driver="]]..query_driver..[["]];
	end;
	if toolset["type"] == "esp-idf" then
		-- remap esp path for compile_commands.json
		local query_driver_src = make_path(edx.app_path, "wsl-query-driver.exe");
		local new_driver_time = utils:file_time(query_driver_src);
		if new_driver_time == nil then
			query_driver_src = make_path(edx.startup_path, "wsl-query-driver.exe");
			new_driver_time = utils:file_time(query_driver_src);
		end;

		local query_driver_target = make_path(build_path, "../target.txt");
		target_file = io.u8_open(query_driver_target, "w+");
		target_file:write(query_driver);
		target_file:close();

		local query_driver = make_path(build_path, "../esp32-none-gcc.exe");
		local old_driver_time = utils:file_time(query_driver);
		if old_driver_time == nil or new_driver_time > old_driver_time then
			if old_driver_time ~= nil then
				os.remove(query_driver);
			end;
			print(LANG("msg/clangd/install_query_driver"), query_driver);
			utils:copy_file(query_driver_src, query_driver);
		end;
		query_driver = query_driver:gsub("\\", "/");
		param_query_driver = [[ --query-driver="]]..query_driver..[["]];
	end;
	lspc:init(
		"ClangD",
		clangd_path,
		[[--pch-storage=memory ]]..
		[[--all-scopes-completion ]]..
		-- [[--completion-style=detailed ]]..
		[[--completion-style=bundled ]]..
		[[--include-ineligible-results ]]..
		[[--j=8 ]]..
		[[--limit-results=200 ]]..
		[[--header-insertion-decorators=0 ]]..
		-- [[--sync ]]..
		-- [[--log=verbose ]]..
		[[--background-index ]]..
		[[--use-dirty-headers ]]..
		param_compile_commands_dir..
		param_query_driver
		, cmd.project_path
		, lsp_env
	);
	lspc.__startup_toolset = toolset;
	lspc.__ext_formatting = utils:file_time(make_path(cmd.project_path, ".clang-format") or "") ~= nil;
end;
function lsp:_init_ccls(cmd)
	local ccls_root = [[D:\Tools\ccls]];
	local ccls_path = make_path(ccls_root, "ccls.exe");
	local lspc = _init_lspc(edx:create_lsp_client());
	local lsp_env = nil;
	edx._lsp_clients["cpp"] = lspc;
	local build_path = cmd.build_path;
	local toolset = cmd.toolset;
	lspc.send_empty_command = LSPC_SEND_EMPTY_NOTIFY;
	local compile_commands = make_path(build_path, "/compile_commands.json");

	local ccls_config = {
		cache = {
			directory = make_path(build_path,".ccls-cache");
		};
		clang = {
			resourceDir = make_path(ccls_root, "clang/14.0.6");
		};
		compilationDatabaseDirectory = build_path;
	};
	local ccls_config_json = utils:tojson(ccls_config);

	lspc:init(
		"CCLS",
		ccls_path,
		[[--init=]]..(ccls_config_json:gsub("\"", [["""]]))
		.. [[ -v=1]]
		-- ..[[ --log-file=]]..make_path(build_path, "/ccls-log.txt")
		, cmd.project_path
		, lsp_env
	);
end;
function lsp:_init_serve_d(cmd)
	local lspc = _init_lspc(edx:create_lsp_client());
	local lsp_env = nil;
	edx._lsp_clients["d"] = lspc;
	lspc:init(
		"Serve-D",
		D_LANGUAGE_SERVER,
		[[]]
		, cmd.project_path
		, lsp_env
	);
end;
function lsp:_init_lua_lang_server(cmd)
	local lspc = _init_lspc(edx:create_lsp_client());
	local lsp_env = nil;
	local project_path = cmd and cmd.project_path;
	-- local project_path = make_path(mgr.current_document.file_name, "../");
	edx._lsp_clients["lua"] = lspc;
	lspc.send_empty_command = LSPC_NO_EMPTY_COMMAND;
	local startup_script = make_path(LUA_LANGUAGE_SERVER, "../../main.lua");
	startup_script = startup_script:gsub("\\", "/");
	lspc:init(
		"Lua-Language-Server",
		LUA_LANGUAGE_SERVER,
		[[-E ]]..startup_script
		, project_path
		, lsp_env
	);
end;
function lsp:_init_lualsp_lang_server(cmd)
	local lspc = _init_lspc(edx:create_lsp_client());
	local lsp_env = nil;
	local project_path = cmd and cmd.project_path;
	edx._lsp_clients["lua"] = lspc;
	lspc.send_empty_command = LSPC_NO_EMPTY_COMMAND;
	lspc:init(
		"LuaLSP",
		LUALSP_LANGUAGE_SERVER,
		[[-localpath ]]..project_path..
		[[ -mode 1]]
		, project_path
		, lsp_env
	);
end;
function lsp:_init_java_lang_server(cmd)
	local lspc = _init_lspc(edx:create_lsp_client());
	local lsp_env = nil;
	local project_path = cmd and cmd.project_path;
	-- local project_path = make_path(mgr.current_document.file_name, "../");
	edx._lsp_clients["java"] = lspc;
	lspc.send_empty_command = LSPC_SEND_EMPTY_NOTIFY;
	lspc:init(
		"JDT-Languate-Server",
		[[java]],
		[[ -Declipse.application=org.eclipse.jdt.ls.core.id1]]..
		[[ -Dosgi.bundles.defaultStartLevel=4]]..
		[[ -Declipse.product=org.eclipse.jdt.ls.core.product]]..
		[[ -Dlog.level=ALL -noverify -Xmx1G]]..
		[[ -jar D:\jdt-language-server\plugins\org.eclipse.equinox.launcher_1.6.400.v20210924-0641.jar]]..
		[[ -configuration D:\jdt-language-server\config_win]]..
		[[ -data D:\jdt-language-server\data]]..
		[[ --add-modules=ALL-SYSTEM]]..
		[[ --add-opens java.base/java.util=ALL-UNNAMED]]..
		[[ --add-opens java.base/java.lang=ALL-UNNAMED]]
		, project_path
		, lsp_env
	);

end;
function lsp:_init_jedi_lang_server(cmd)
	local lspc = _init_lspc(edx:create_lsp_client());
	local lsp_env = nil;
	local project_path = cmd and cmd.project_path;
	-- local project_path = make_path(mgr.current_document.file_name, "../");
	edx._lsp_clients["python"] = lspc;
	lspc.send_empty_command = LSPC_NO_EMPTY_COMMAND;
	lspc:init(
		"Jedi-Languate-Server",
		JEDI_LANGUAGE_SERVER,
		[[]]
		, project_path
		, lsp_env
	);
end;
function lsp:_init_pylsp_lang_server(cmd)
	local lspc = _init_lspc(edx:create_lsp_client());
	local lsp_env = nil;
	local project_path = cmd and cmd.project_path;
	-- local project_path = make_path(mgr.current_document.file_name, "../");
	edx._lsp_clients["python"] = lspc;
	lspc.send_empty_command = LSPC_NO_EMPTY_COMMAND;
	lspc:init(
		"Python-Languate-Server",
		PYLSP_LANGUAGE_SERVER,
		[[]]
		, project_path
		, lsp_env
	);
end;
function lsp:_init_cmake_lang_server(cmd)
	local lspc = _init_lspc(edx:create_lsp_client());
	local lsp_env = nil;
	local project_path = cmd and cmd.project_path;
	-- local project_path = make_path(mgr.current_document.file_name, "../");
	edx._lsp_clients["cmake"] = lspc;
	lspc.send_empty_command = LSPC_NO_EMPTY_COMMAND;
	lspc:init(
		"CMake-Languate-Server",
		CMAKE_LANGUAGE_SERVER,
		[[--stdio]]
		, project_path
		, lsp_env
	);
end;
function lsp:_completion_exec(doc, func)
	if not edx._lsp_clients then
		return false;
	end;
	local content_type_id = content_type_name[doc.content_type];
	local lspc = edx._lsp_clients[content_type_id];
	if not lspc then
		return false;
	end;
	return dbg_call(function()
			return func(lspc);
	end, "lsp error");
end;
function lsp:_try_init_lsp(working_dir)
	if edx._lsp_clients then
		return false;
	end;
	local cmd = project:find_build_command();
	if cmd and cmd["type"] ~= "direct" then
		return false;
	end;
	if not working_dir then
		working_dir = "";
	end;
	if not cmd then
		cmd = {
			build_path = "";
			toolset = TOOLSETS[1];
			project_path = working_dir;
		};
	end;
	if not cmd.toolset then
		cmd.toolset = TOOLSETS[1];
	end;
	if not cmd.build_path then
		cmd.build_path = "";
	end;
	if not cmd.project_path then
		cmd.project_path = working_dir;
	end;
	self:reset(cmd);
	return true;
end;
function lsp:is_lspc_ready(lang)
	return edx._lsp_clients and edx._lsp_clients[lang];
end;
function lsp:terminate()
	if not edx._lsp_clients then
		return;
	end;
	for name, lspc in pairs(edx._lsp_clients) do
		lspc:terminate();
		lspc.__startup_toolset = nil;
	end;

	edx._lsp_clients = nil;
	edx._prepare_lsp = nil;
end;
function lsp:reset(cmd)
	self:terminate();
	edx._lsp_clients = {};
	edx:clear("lsp");
	edx._prepare_lsp = function(lang)
		-- print("Prepare LSP:", lang, edx._lsp_clients);
		if edx._lsp_clients[lang] then
			return;
		end;
		if lang == "cpp" then
			-- init clangd service
			self:_init_clangd(cmd);
		elseif lang == "python" then
			if PYLSP_LANGUAGE_SERVER then
				-- init pylsp service
				self:_init_pylsp_lang_server(cmd);
			elseif JEDI_LANGUAGE_SERVER then
				-- init jedi service
				self:_init_jedi_lang_server(cmd);
			end;
		elseif lang == "cmake" then
			if CMAKE_LANGUAGE_SERVER then
				-- init cmake service
				self:_init_cmake_lang_server(cmd);
			end;
		elseif lang == "lua" then
			if LUA_LANGUAGE_SERVER then
				-- init lua-language-server service
				self:_init_lua_lang_server(cmd);
			elseif LUALSP_LANGUAGE_SERVER then
				-- init lualsp service
				self:_init_lualsp_lang_server(cmd);
			end;
		elseif lang == "d" then
			if D_LANGUAGE_SERVER then
				-- init serve-d service
				self:_init_serve_d(cmd);
			end;
		end;
	end;
	-- self:_init_ccls(cmd);

	-- init serve-d service
	-- self:_init_serve_d(cmd);

	-- self:_init_lua_lang_server(cmd);

	-- self:_init_java_lang_server(cmd);

	local all_doc = {mgr.get_all_documents()};
	local def_target = menu_bar.get_cmake_default_target();
	for idx, doc in ipairs(all_doc) do
		local content_type_id = content_type_name[doc.content_type];
		edx._prepare_lsp(content_type_id);
		local lspc = edx._lsp_clients[content_type_id];
		if lspc then
			doc.ext_formatting = lspc.__ext_formatting;
			lspc:did_open_doc(doc, content_type_id);
		else
			doc.ext_formatting = false;
		end;
	end;
end;
---@param doc document
function lsp:get_startup_toolset(doc)
	if not edx._lsp_clients then
		return nil;
	end;
	local doc_type = doc;
	if type(doc_type) ~= "number" then
		doc_type = doc.content_type;
	end;
	local content_type_id = content_type_name[doc_type];
	local lspc = edx._lsp_clients[content_type_id];
	if lspc then
		return lspc.__startup_toolset;
	end;
	return nil;
end;
---@param doc document
function lsp:doc_saved(doc)
	self:_completion_exec(doc, function(lspc) return lspc:did_save_doc(doc); end);
end;
---@param doc document
function lsp:doc_opened(doc)
	-- try init lsp for direct mode
	if self:_try_init_lsp() then
		-- initialized, all opened document will invoke did_open_doc during reseting
		return;
	end;
	local content_type_id = content_type_name[doc.content_type];
	if edx._prepare_lsp then
		edx._prepare_lsp(content_type_id);
	end;
	local lspc = edx._lsp_clients and edx._lsp_clients[content_type_id];
	if lspc then
		doc.ext_formatting = lspc.__ext_formatting;
	else
		doc.ext_formatting = false;
	end;
	self:_completion_exec(doc, function(lspc) return lspc:did_open_doc(doc, content_type_id); end);
end;
---@param doc document
function lsp:doc_closed(doc)
	self:_completion_exec(doc, function(lspc) return lspc:did_close_doc(doc); end);
end;
---@param doc document
function lsp:doc_renamed(doc, from_path, to_path)
	self:_completion_exec(doc, function(lspc) return lspc:did_rename_doc(doc, from_path, to_path); end);
end;
---@param doc document
function lsp:find_signature_help(doc, helper)
	if self:_completion_exec(doc, function(lspc) return lspc:find_signature_help(doc, helper) or (helper.cancel_completion() and false); end) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:do_completion(doc, helper)
	if self:_completion_exec(doc, function(lspc) return lspc:do_complete(doc, helper) or (helper and helper.cancel_completion() and false); end) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:update_completion_tip(doc, helper)
	if self:_completion_exec(doc, function(lspc) return lspc:update_completion_list_tip(doc, helper) or (helper.cancel_completion() and false); end) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:cleanup_completion(doc)
	if edx._lsp_clients then
		if doc then
			local content_type_id = content_type_name[doc.content_type];
			local lspc = edx._lsp_clients[content_type_id];
			if lspc then
				lspc:clean_completion_list();
			end;
		else
			for idx, lspc in ipairs(edx._lsp_clients) do
				lspc:clean_completion_list();
			end;
		end;
	end;
end;
-- commit_mode flags
-- 0x01 : extend mode
-- 0x02 : commit additional edit
---@param doc document
function lsp:commit_completion(doc, helper, commit_mode)
	return self:_completion_exec(doc, function(lspc) return lspc:commit_completion(doc, helper, commit_mode); end);
end;
---@param doc document
function lsp:hover_info(doc, line, col, pop_by_cursor)
	if line == nil then
		line = -1;
	end;
	if col == nil then
		col = -1;
	end;
	if pop_by_cursor == true then
		-- popup by cursor & auto hide
		pop_by_cursor = 1 + 0x8000;
	else
		pop_by_cursor = 0;
	end;
	if(self:_completion_exec(doc, function(lspc) return lspc:find_hover_info(doc, line, col, pop_by_cursor); end)) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:find_declaration(doc)
	if(self:_completion_exec(doc, function(lspc) return lspc:find_declaration(doc); end)) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:find_definition(doc)
	if(self:_completion_exec(doc, function(lspc) return lspc:find_definition(doc); end)) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:find_implementation(doc)
	if(self:_completion_exec(doc, function(lspc) return lspc:find_implementation(doc); end)) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:find_reference(doc)
	if(self:_completion_exec(doc, function(lspc) return lspc:find_reference(doc); end)) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:list_doc_structure(doc, class_view)
	if(self:_completion_exec(doc, function(lspc) return lspc:list_doc_structure(doc, class_view); end)) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:find_code_action(doc, types)
	if(self:_completion_exec(doc, function(lspc) return lspc:find_code_action(doc, types); end)) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:cancel_completion(doc)
	if(self:_completion_exec(doc, function(lspc) return lspc:cancel_completion(doc); end)) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:cancel_signature_help(doc)
	if(self:_completion_exec(doc, function(lspc) return lspc:cancel_signature_help(doc); end)) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:cancel_list_doc_structure(doc)
	if(self:_completion_exec(doc, function(lspc) return lspc:cancel_list_doc_structure(doc); end)) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:do_on_type_formatting(doc)
	if(self:_completion_exec(doc, function(lspc) return lspc:do_on_type_formatting(doc); end)) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:do_range_formatting(doc, begin_line, begin_col, end_line, end_col)
	if(self:_completion_exec(doc, function(lspc) return lspc:do_range_formatting(doc, begin_line, begin_col, end_line, end_col); end)) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:do_formatting(doc)
	if(self:_completion_exec(doc, function(lspc) return lspc:do_formatting(doc); end)) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:rename_symbol(doc, new_name)
	if(self:_completion_exec(doc, function(lspc) return lspc:rename_symbol(doc, new_name); end)) then
		return true;
	end;
	return false;
end;
---@param doc document
function lsp:find_inactive_regions(doc)
	if(self:_completion_exec(doc, function(lspc) return lspc:find_inactive_regions(doc); end)) then
		return true;
	end;
	return false;
end;
